#include <windows.h>
#include <shlobj.h>
#include "out.h"  // winamp header
#include <vorbis/vorbisfile.h>
#include <vorbis/modes.h>

#define PI_VER2 "v1.0"

#ifdef __alpha
#define PI_VER PI_VER2 " (AXP)"
#else
#define PI_VER PI_VER2 " (x86)"
#endif

static char app_name[] = "OGG Writing Output Driver";
static char INI_FILE[MAX_PATH];
extern char config_waveoutdir[MAX_PATH];

int getwrittentime();
void config_read();
void config_write();

char config_waveoutdir[MAX_PATH]="";
FILE *mtdout;
int srate, numchan, bps;
volatile int writtentime, w_offset;
Out_Module out;
static int last_pause=0;

ogg_stream_state os; /* take physical pages, weld into a logical stream of packets */
ogg_page         og; /* one Ogg bitstream page.  Vorbis packets are inside */
ogg_packet       op; /* one raw packet of data for decode */
vorbis_info     *vi; /* struct that stores all the static vorbis bitstream settings */
vorbis_comment   vc; /* struct that stores all the user comments */
vorbis_dsp_state vd; /* central working state for the packet->PCM decoder */
vorbis_block     vb; /* local working space for packet->PCM de*/

int eos=0;

#define READ 1024
signed char readbuffer[READ*4+44]; /* out of the data segment, not the stack */

static int CALLBACK WINAPI BrowseCallbackProc( HWND hwnd, UINT uMsg, LPARAM lParam, LPARAM lpData)
{
    if (uMsg == BFFM_INITIALIZED)
    {
        SetWindowText(hwnd,"Select Directory");
        SendMessage(hwnd,BFFM_SETSELECTION,(WPARAM)1,(LPARAM)config_waveoutdir);
    }
    return 0;
}

BOOL WINAPI _DllMainCRTStartup(HANDLE hInst, ULONG ul_reason_for_call, LPVOID lpReserved)
{
    if (ul_reason_for_call == DLL_PROCESS_ATTACH)
        DisableThreadLibraryCalls(hInst);
    return TRUE;
}

void config(HWND hwnd)
{
    char name[MAX_PATH];
    BROWSEINFO bi;
    ITEMIDLIST *idlist;
    bi.hwndOwner = hwnd;
    bi.pidlRoot = 0;
    bi.pszDisplayName = name;
    bi.lpszTitle = "Select a directory for Ogg Vorbis file output:";
    bi.ulFlags = BIF_RETURNONLYFSDIRS;
    bi.lpfn = BrowseCallbackProc;
    bi.lParam = 0;
    idlist = SHBrowseForFolder( &bi );
    if (idlist)
    {
        SHGetPathFromIDList( idlist, config_waveoutdir );
        config_write();
    }
}

void about(HWND hwnd)
{
    MessageBox(hwnd,"Ogg Vorbis Writer Plug-in " PI_VER "\n"
               "Copyright (C)2000 ???? GPL\n\n"
               "Compiled on " __DATE__ "\n","About",MB_OK);
}

void init()
{
    config_read();
}

void quit()
{

}

char *scanstr_back(char *str, char *toscan, char *defval)
{
    char *s=str+strlen(str)-1;
    if (strlen(str) < 1)
        return defval;
    if (strlen(toscan) < 1)
        return defval;
    while (1)
    {
        char *t=toscan;
        while (*t)
            if (*t++ == *s)
                return s;
        t=CharPrev(str,s);
        if (t==s)
            return defval;
        s=t;
    }
}

int open(int samplerate, int numchannels, int bitspersamp, int bufferlenms, int prebufferms)
{
    char *t,*p, combuf[512]; // hard-coded buffer!
    char temp2[MAX_PATH],temp[MAX_PATH];
    ogg_packet header;
    ogg_packet header_comm;
    ogg_packet header_code;

    // Straight from the winamp example -
    // A lot of crap just to make a filename. :)
    // I am gonna change this REAL soon. It makes the
    // ogg comments ugly, and the filename unix-unfriendly
    GetWindowText(out.hMainWindow,temp2,sizeof(temp2));
    t=temp2;

    t=scanstr_back(temp2,"-",NULL);
    if (t) t[-1]=0;

    if (temp2[0] && temp2[1] == '.')
    {
        char *p1,*p2;

        p1=temp;
        p2=temp2;
        while (*p2)
            *p1++=*p2++;
        *p1=0;
        p1 = temp2+1;
        p2 = temp;
        while (*p2)
        {
            *p1++ = *p2++;
        }
        *p1=0;
        temp2[0] = '0';
    }
    p=temp2;
    while (*p != '.' && *p)
        p++;
    if (*p == '.')
    {
        *p = '-';
        p=CharNext(p);
    }
    while (*p)
    {
        if (*p == '.' || *p == '/' || *p == '\\' || *p == '*' ||
            *p == '?' || *p == ':' || *p == '+' || *p == '\"' ||
            *p == '\'' || *p == '|' || *p == '<' || *p == '>') *p = '_';
        p=CharNext(p);
    }

    p=config_waveoutdir;
    if (p[0])
        while (p[1])
            p++;

    if (!config_waveoutdir[0] || config_waveoutdir[0] == ' ')
        config(out.hMainWindow);
    if (!config_waveoutdir[0])
        wsprintf(temp,"%s.ogg",temp2);
    else if (p[0]=='\\')
        wsprintf(temp,"%s%s.ogg",config_waveoutdir,temp2);
    else
        wsprintf(temp,"%s\\%s.ogg",config_waveoutdir,temp2);


    // We are gonna ignore this and hard-code an encoding mode?
    w_offset = writtentime = 0;
    numchan = numchannels;
    srate = samplerate;
    bps = bitspersamp;

    // The fp for the output file. I know it's not windows style. So?
    mtdout = fopen(temp, "wb");
    if(mtdout==NULL)
        return -1;

    /********** Encode setup ************/

    /* choose an encoding mode */
    /* (mode 0: 44kHz stereo uncoupled, roughly 128kbps VBR) */
    vi=&info_A;

    /* add a comment */
    vorbis_comment_init(&vc);
    sprintf(&combuf[0], "%s (WinAmp encoded OGG %s)", temp2, __DATE__);
    vorbis_comment_add(&vc,combuf);

    /* set up the analysis state and auxiliary encoding storage */
    vorbis_analysis_init(&vd, vi);
    vorbis_block_init(&vd,&vb);

    /* set up our packet->stream encoder */
    /* pick a random serial number; that way we can more likely build
     chained streams just by concatenation */
    srand(GetTickCount());
    ogg_stream_init(&os,rand());

    /* Vorbis streams begin with three headers; the initial header (with
     most of the codec setup parameters) which is mandated by the Ogg
     bitstream spec.  The second header holds any comment fields.  The
     third header holds the bitstream codebook.  We merely need to
     make the headers, then pass them to libvorbis one at a time;
     libvorbis handles the additional Ogg bitstream constraints */

    vorbis_analysis_headerout(&vd,&vc,&header,&header_comm,&header_code);
    ogg_stream_packetin(&os,&header); /* automatically placed in its own page */
    ogg_stream_packetin(&os,&header_comm);
    ogg_stream_packetin(&os,&header_code);

    return 0;
}

void close()
{
    ogg_stream_clear(&os);
    vorbis_block_clear(&vb);
    vorbis_dsp_clear(&vd);

    if(mtdout)
        fclose(mtdout);
}

int write(char *buf, int len)
{
    int i;
    double **buffer;
    writtentime += len;

    memcpy(readbuffer, buf, len);

    /* expose the buffer to submit data */
    buffer=vorbis_analysis_buffer(&vd,READ);

    /* uninterleave samples */
    for(i=0;i<len/4;i++)
    {
        buffer[0][i]=((readbuffer[i*4+1]<<8)|
                      (0x00ff&(int)readbuffer[i*4]))/32768.;
        buffer[1][i]=((readbuffer[i*4+3]<<8)|
                      (0x00ff&(int)readbuffer[i*4+2]))/32768.;
    }

    /* tell the library how much we actually submitted */
    vorbis_analysis_wrote(&vd,i);


    /* vorbis does some data preanalysis, then divvies up blocks for
     more involved (potentially parallel) processing.  Get a single
     block for encoding now */
    while(vorbis_analysis_blockout(&vd,&vb)==1)
    {
        /* analysis */
        vorbis_analysis(&vb,&op);

        /* weld the packet into the bitstream */
        ogg_stream_packetin(&os,&op);

        /* write out pages (if any) */
        while(!eos)
        {
            int result=ogg_stream_pageout(&os,&og);
            if(result==0)
                break;
            fwrite(og.header,1,og.header_len,mtdout);
            fwrite(og.body,1,og.body_len,mtdout);

            if(ogg_page_eos(&og))
                eos=1;
        }
    }

    Sleep(0); // This was in the example plugin. dunno.
    return 0;
}

int canwrite()
{
    return last_pause?0:4*1024;
    //return last_pause?0:16*1024*1024;
}

int isplaying()
{
    return 0;
}

int pause(int pause)
{
    int t=last_pause;
    last_pause=pause;
    return t;
}

void setvolume(int volume)
{

}

void setpan(int pan)
{

}

void flush(int t)
{
    int a;
    w_offset=0;
    a = t - getwrittentime();
    w_offset=a;
}

int getwrittentime()
{
    int t=srate*numchan,l;
    int ms=writtentime;

    l=ms%t;
    ms /= t;
    ms *= 1000;
    ms += (l*1000)/t;

    if (bps == 16) ms/=2;

    return ms + w_offset;
}

Out_Module out =
{
    OUT_VER,
    "Ogg Vorbis Writer plug-in " PI_VER,
    33,
    0, // hmainwindow
    0, // hdllinstance
    config,
    about,
    init,
    quit,
    open,
    close,
    write,
    canwrite,
    isplaying,
    pause,
    setvolume,
    setpan,
    flush,
    getwrittentime,
    getwrittentime
};

static void _r_s(char *name,char *data, int mlen)
{
    char buf[2048];
    strcpy(buf,data);
    GetPrivateProfileString(app_name,name,buf,data,mlen,INI_FILE);
}

#define RS(x) (_r_s(#x,x,sizeof(x)))
#define WS(x) (WritePrivateProfileString(app_name,#x,x,INI_FILE))

static void config_init()
{
    char *p=INI_FILE;
    GetModuleFileName(NULL,INI_FILE,sizeof(INI_FILE));
    while (*p)
        p++;
    while (p >= INI_FILE && *p != '.')
        p--;
    strcpy(p+1,"ini");
}

void config_read()
{
    config_init();
    RS(config_waveoutdir);
}

void config_write()
{
    WS(config_waveoutdir);
}

__declspec( dllexport ) Out_Module * winampGetOutModule()
{
    return &out;
}

